unit fRichEditControlElements;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Buttons, StdCtrls, ComCtrls;

type
  TFrameRichEditControlElements = class(TForm)
    labCurrentLine: TLabel;
    labLineNr: TLabel;
    butGotoLine: TSpeedButton;
    butSearch: TSpeedButton;
    FindDialog: TFindDialog;
    procedure FormCreate(Sender: TObject);
    procedure FindDialogFind(Sender: TObject);
    procedure butSearchClick(Sender: TObject);
    procedure butGotoLineClick(Sender: TObject);
  private
    RichEdit: TRichEdit;
    Procedure RESelectionChange(Sender: TObject);
  public
    Procedure StartSearch;
  end;

Function UseRichEditControlElements(Const RE: TRichEdit;
   Const AutoShow: Boolean = false): TFrameRichEditControlElements;
  
implementation

{$R *.DFM}

Function UseRichEditControlElements(Const RE: TRichEdit;
   Const AutoShow: Boolean = false): TFrameRichEditControlElements;
begin
   Result := TFrameRichEditControlElements.Create(Re.Owner);
   With Result do try
      Parent := Re.Parent;
      ClientHeight := butGotoLine.Height + 2 * butGotoLine.Top;
      Richedit := Re;
      Richedit.OnSelectionChange := RESelectionChange;
      RichEdit.HideSelection := false;
      RESelectionChange(RichEdit);
      FindDialog.Options := [frDown, frHideUpDown];
      Visible := AutoShow
   except
      Free;
      Result := NIL
   end
end;

procedure TFrameRichEditControlElements.RESelectionChange(Sender: TObject);
begin
   With Richedit do begin
      labLineNr.Caption := IntToStr(SendMessage(Handle, EM_LINEFROMCHAR, SelStart, 0)+1)
   end
end;

procedure TFrameRichEditControlElements.FormCreate(Sender: TObject);
begin
   Name := ''; Caption := ''
end;

procedure TFrameRichEditControlElements.FindDialogFind(Sender: TObject);
Var FoundAt: LongInt;
    StartPos, ToEnd: Integer;
    Searchtype: TSearchTypes;
begin
   with RichEdit do begin
      if SelLength <> 0
         then StartPos := SelStart + SelLength
         else StartPos := 0;
      Searchtype := [];
      With FindDialog do begin
         If Not (frFindNext IN Options) then StartPos := 0;
         If frMatchCase IN Options then Searchtype := Searchtype + [stMatchCase];
         If frWholeWord IN Options then Searchtype := Searchtype + [stWholeWord];
      end;
      ToEnd := Length(Text) - StartPos;
      FoundAt := FindText(FindDialog.FindText, StartPos, ToEnd, Searchtype);
      If (FoundAt < 0) and (StartPos > 0) then FoundAt := FindText(FindDialog.FindText, 0, Length(Text), Searchtype);
      if FoundAt = -1 then begin
         Beep
      end else begin
         SelStart := FoundAt;
         SelLength := Length(FindDialog.FindText)
      end
   end
end;

procedure TFrameRichEditControlElements.butSearchClick(Sender: TObject);
begin
   StartSearch
end;

procedure TFrameRichEditControlElements.butGotoLineClick(Sender: TObject);
Var s: String;
begin
   s := labLineNr.Caption;
   If InputQuery('Goto line', 'New line number?', s) then begin
      With Richedit do SelStart := SendMessage(Handle, EM_LINEINDEX, StrToInt(s)-1, 0)
   end
end;

procedure TFrameRichEditControlElements.StartSearch;
begin
   FindDialog.Execute
end;

end.
